"""Core trufflehog3 logic."""

import attr
import multiprocessing
import sys
import yaml

from functools import partial
from itertools import chain
from pathlib import Path
from typing import Iterable, List, Union

from trufflehog3 import DEFAULT_CONFIG_FILE, DEFAULT_RULES_FILE
from trufflehog3 import log
from trufflehog3.models import (
    Config,
    Entropy,
    Format,
    Issue,
    Model,
    Pattern,
    Rule,
    Severity,
)
from trufflehog3.render import text, json, html
from trufflehog3.search import search
from trufflehog3.source import diriter, gititer


def scan(
    target: str,
    config: Config,
    rules: Iterable[Union[Entropy, Pattern]],
    processes: int,
) -> Iterable[Issue]:
    """Return issues found during target path scan."""
    exclude = []
    for e in config.exclude or []:
        if e.id is None and e.pattern is None:
            exclude.extend(e.paths)

    files = []
    if not config.no_history:  # pragma: no cover
        files.extend(
            gititer(
                target,
                exclude=exclude,
                branch=config.branch,
                depth=config.depth,
                since=config.since,
            )
        )

    if not config.no_current:  # pragma: no cover
        files.extend(diriter(target, exclude))

    worker = partial(
        search,
        rules=rules,
        exclude=config.exclude,
        ignore_nosecret=config.ignore_nosecret,
        context=config.context,
    )

    with multiprocessing.Pool(processes) as pool:
        issues = pool.map(worker, files)

    return set(chain.from_iterable(issues))


def diff(
    old: Iterable[Issue],
    new: Iterable[Issue],
    only_new: bool = False,
) -> Iterable[Issue]:  # pragma: no cover
    """Return diff from the given lists of issues.

    By default, full diff will be returned, i.e. issues that are not found in
    either of the given lists. In case `only_new` is True, only issues that
    are present in the `new` list, but not in the `old` one will be returned.

    """
    old_set = set(old)
    new_set = set(new)

    if only_new:
        d = new_set - old_set
    else:
        d = new_set ^ old_set

    return list(d)


def load_config(path: str, **kwargs) -> Config:
    """Load config from file or search for it in the specified directory.

    Note
    ----
    Uses `DEFAULT_CONFIG_FILE` filename for searching config file in directory.

    All `kwargs` are passed to `Config.update` method after loading.

    Examples
    --------
    Basic usage examples. Load config from file

    >>> path = Path() / DEFAULT_CONFIG_FILE
    >>> config = load_config(path)
    >>> config.ignore_nosecret
    False

    Provide values to override config

    >>> config = load_config(path, ignore_nosecret=True)
    >>> config.ignore_nosecret
    True

    Search config in directory

    >>> config = load_config(Path())
    >>> config.ignore_nosecret
    False

    """
    path = Path(path)
    for config_path in (path, path / DEFAULT_CONFIG_FILE):
        if config_path.is_file():
            log.info(f"loading config from {config_path.absolute()}")
            config = load(Config, config_path)
            break
    else:
        config = Config()  # pragma: no cover

    return attr.evolve(config, **kwargs)


def load_rules(
    path: str,
    severity: Severity = Severity.LOW,
) -> Iterable[Union[Entropy, Pattern]]:
    """Load rules from file.

    Examples
    --------
    Basic usage examples

    >>> len(load_rules(DEFAULT_RULES_FILE))
    31
    >>> len(load_rules(DEFAULT_RULES_FILE, Severity.MEDIUM))
    28
    >>> len(load_rules(DEFAULT_RULES_FILE, Severity.HIGH))
    2
    """
    rules = []
    for r in load(Rule.fromargs, path or DEFAULT_RULES_FILE):
        if r.severity >= severity:
            rules.append(r)
        else:
            log.warning("skipping rule: %s", r.id)

    return rules


def load(cls: type, file: str = None) -> Union[Model, List[Model]]:
    """Load any model from YAML file.

    Note
    ----
    File defaults to `sys.stdin`. See `core.loads` for more details.

    Examples
    --------
    Define a class with the necessary properties

    >>> @attr.s
    ... class Test(Model):
    ...     version: str = attr.ib()
    ...     numbers: List[int] = attr.ib()

    Load model

    >>> obj = load(Test, "tests/data/test_load.yml")
    >>> obj.asdict()
    {'version': 'v1', 'numbers': [1, 2]}

    Load list of models

    >>> objs = load(Test, "tests/data/test_load_list.yml")
    >>> [obj.asdict() for obj in objs]
    [{'version': 'v1', 'numbers': [1, 2]}, {'version': 'v2', 'numbers': [3]}]

    """
    if file:
        raw = Path(file).read_text()
    else:
        raw = sys.stdin.read()  # pragma: no cover

    return loads(cls, raw)


def loads(cls: type, raw: str) -> Union[Model, List[Model]]:
    """Load any model from YAML string.

    Note
    ----
    Model is instantiated using provided `cls`.
    If YAML contains a list, each item is instantiated using `cls`

    Examples
    --------
    Define a class with the necessary properties

    >>> @attr.s
    ... class Test(Model):
    ...     version: str = attr.ib()
    ...     numbers: List[int] = attr.ib()

    Load model

    >>> raw = '''
    ... version: v1
    ... numbers:
    ... - 1
    ... - 2
    ... '''
    >>> obj = loads(Test, raw)
    >>> obj.asdict()
    {'version': 'v1', 'numbers': [1, 2]}

    Load list of models

    >>> raw = '''
    ... - version: v1
    ...   numbers:
    ...   - 1
    ...   - 2
    ... - version: v2
    ...   numbers:
    ...   - 3
    ... '''
    >>> objs = loads(Test, raw)
    >>> [obj.asdict() for obj in objs]
    [{'version': 'v1', 'numbers': [1, 2]}, {'version': 'v2', 'numbers': [3]}]

    """
    data = yaml.safe_load(raw)
    if isinstance(data, list):
        model = [cls(**item) for item in data]
    else:
        model = cls(**data)

    return model


def render(issues: Iterable[Issue], format: Format, file: str = None):
    """Render issues to file in given format."""
    write(renders(issues, format), file)


def renders(
    issues: Iterable[Issue],
    format: Format,
) -> str:  # pragma: no cover
    """Render issues to string in given format."""
    if format == Format.TEXT:
        f = text
    elif format == Format.JSON:
        f = json
    elif format == Format.HTML:
        f = html
    else:
        raise ValueError(f"unknown render format: '{format}'")

    return f(issues)


def dump(obj: Union[Model, List[Model]], file: str = None):  # pragma: no cover
    """Dump model to file.

    Notes
    -----
    File defaults to `sys.stdout`. See `core.dumps` for more details.

    """
    write(dumps(obj), file)


def dumps(model: Union[Model, List[Model]], **kwargs) -> str:
    """Dump model to string.

    Note
    ----
    Model is serialized to YAML using its own `asdict` method.
    If model is a list, each item is serialized using its own `asdict` method.

    All `kwargs` are directly passed to `yaml.safe_dump`.

    Examples
    --------
    Define a class with the necessary properties

    >>> @attr.s
    ... class Test(Model):
    ...     version: str = attr.ib()
    ...     numbers: List[int] = attr.ib()

    Dump model

    >>> obj = Test(version="v1", numbers=[1, 2])
    >>> print(dumps(obj, sort_keys=False), end="")
    version: v1
    numbers:
    - 1
    - 2

    Dump list of models

    >>> objs = [
    ...     Test(version="v1", numbers=[1, 2]),
    ...     Test(version="v2", numbers=[3]),
    ... ]
    >>> print(dumps(objs, sort_keys=False), end="")
    - version: v1
      numbers:
      - 1
      - 2
    - version: v2
      numbers:
      - 3

    """
    if isinstance(model, list):
        data = [item.asdict() for item in model]
    else:
        data = model.asdict()

    return yaml.safe_dump(data, default_flow_style=False, **kwargs)


def write(raw: str, file: str = None):  # pragma: no cover
    """Write string to file.

    Note
    ----
    File defaults to `sys.stdout`.

    """
    if file:
        Path(file).write_text(raw)
    else:
        sys.stdout.write(raw)



"""Supported search sources."""

import git
import os

from pathlib import Path
from typing import Iterable, Iterator, Optional

from trufflehog3 import DEFAULT_EXCLUDE_SET
from trufflehog3 import log
from trufflehog3.models import File


def dirlist(path: str, exclude: Iterable[str] = None) -> Iterable[File]:
    """Recursively iterate over directory and return existing files.

    Examples
    --------
    Basic usage examples

    >>> for d in dirlist("trufflehog3/static", exclude=["*.yml"]):
    ...     print(d.path)
    report.text.j2
    report.html.j2

    """
    return list(diriter(path, exclude))


def diriter(path: str, exclude: Iterable[str] = None) -> Iterator[File]:
    """Recursively iterate over directory and yield existing files."""
    exclude_set = DEFAULT_EXCLUDE_SET | set(exclude or [])
    # Using `os.walk` here since it allows to drop whole directories.
    # `Path.rglob` requires checking every file against exclude rules.
    # This helps to save a lot of time when excluding large directories.
    for dirpath, dirnames, filenames in os.walk(path):
        rel = Path(dirpath).relative_to(path)

        for directory in dirnames:
            dirname = rel / directory
            pattern = _match(dirname, exclude_set)
            if pattern:
                log.debug(f"skipping directory '{dirname}': '{pattern}'")
                dirnames.remove(directory)

        for file in filenames:
            filename = rel / file
            if filename.is_symlink():  # pragma: no cover
                continue

            pattern = _match(filename, exclude_set)
            if pattern:
                log.debug(f"skipping file '{filename}': '{pattern}'")
                continue

            yield File(
                path=filename.as_posix(),
                absolute=filename.absolute().as_posix(),
            )


def gitlist(
    path: str,
    exclude: Iterable[str] = None,
    branch: str = None,
    depth: int = None,
    since: str = None,
) -> Iterable[File]:
    """Iterate over Git commit history and return diff blobs for each file.

    Examples
    --------
    Basic usage examples

    >>> files = gitlist(
    ...     ".",
    ...     exclude=["*.toml"],
    ...     branch="1.x",
    ...     since="9e404e6c59d286645b2465aacaf61108ebc12a3a",
    ... )
    >>> for d in files[:3]:
    ...     print(d.path)
    README.md
    LICENSE
    README.md

    """
    return list(gititer(path, exclude, branch, depth, since))


def gititer(
    path: str,
    exclude: Iterable[str] = None,
    branch: str = None,
    depth: int = None,
    since: str = None,
) -> Iterator[File]:
    """Iterate over Git commit history and yield diff blobs for each file."""
    try:
        repo = git.Repo(path)
    except Exception:  # pragma: no cover
        log.warning("not a Git repository: %s", path)
        raise StopIteration

    already_searched = set()
    for branch in _get_branches(repo, branch):
        log.info(f"switching to branch '{branch}'")
        prev_commit = None
        since_reached = False
        commits = repo.iter_commits(branch, max_count=depth)

        for curr_commit in commits:
            if curr_commit.hexsha == since:
                since_reached = True

            if since and since_reached:
                prev_commit = curr_commit
                continue

            diff_id = str(prev_commit) + str(curr_commit)
            if not prev_commit or diff_id in already_searched:
                prev_commit = curr_commit
                continue

            diff = prev_commit.diff(curr_commit, create_patch=True)
            already_searched.add(diff_id)
            yield from _diffiter(diff, prev_commit, branch, exclude)
            prev_commit = curr_commit

        diff = curr_commit.diff(git.NULL_TREE, create_patch=True)
        yield from _diffiter(diff, prev_commit, branch, exclude)


def _diffiter(
    diff: git.DiffIndex,
    commit: git.Commit,
    branch: git.Head,
    exclude: Iterable[str] = None,
) -> Iterator[File]:
    r"""Iterate over commit blobs and yield diffs for each file.

    Note
    ----
    Blobs with change type "D" (delete) and "R" (rename) are skipped.

    Examples
    --------
    Basic usage examples

    >>> repo = git.Repo()
    >>> b1 = b"\x87\x02\xc5\x00p\xf8\x84\x04\xd5bLJ\xec\xe69\xba\x85\xf1V\x81"
    >>> b2 = b"o\xac\x8e\xf5\x88\x8e/\xacM\x0cd^\xcd\xc0\xb1\x9dY$`\x80"
    >>> c1 = git.Commit(repo, b1)
    >>> c2 = git.Commit(repo, b2)
    >>> diff = c1.diff(c2, create_patch=True)
    >>> branch = _get_branches(repo, "master")[0]
    >>> len(list(_diffiter(diff, c1, branch, ["tests/*"]))) == 6
    True

    """
    exclude_set = DEFAULT_EXCLUDE_SET | set(exclude or [])
    for blob in diff:
        if blob.deleted_file or blob.renamed_file:  # pragma: no cover
            continue

        pdiff = blob.diff.decode("utf-8", errors="replace")
        fpath = blob.b_path if blob.b_path else blob.a_path

        if pdiff.startswith("Binary files"):  # pragma: no cover
            continue

        pattern = _match(fpath, exclude_set, recursive=True)
        if pattern:
            log.debug(f"skipping diff '{fpath}': '{pattern}'")
            continue

        yield File(
            path=fpath,
            content=pdiff,
            branch=branch.name,
            message=commit.message.strip(),
            author=f"{commit.author.name} <{commit.author.email}>",
            commit=commit.hexsha,
            date=commit.committed_datetime.isoformat(),
        )


def _get_branches(
    repo: git.Repo, branch: str = None
) -> Iterable[git.Commit]:  # pragma: no cover
    """Return a list of repository branches.

    Try to fetch branches from remote first.
    In case of failure, use `branch` as a fallback.
    Otherwise, use active repository branch.

    Examples
    --------
    Basic usage examples

    >>> repo = git.Repo()
    >>> len(_get_branches(repo)) > 1
    True
    >>> _get_branches(repo, "master")[0].name
    'FETCH_HEAD'

    """
    if repo.remotes:
        try:
            return repo.remotes.origin.fetch(branch)
        except git.GitCommandError as e:
            log.warning(f"fetching remote branches: {e}")
    else:
        log.warning("missing remotes")

    return [repo.branches[branch] if branch else repo.active_branch]


def _match(
    path: str, patterns: Iterable[str] = None, recursive: bool = False
) -> Optional[str]:
    """Match path against given glob patterns and return matched pattern if any.

    If `recursive` is set to True, check all parent directories as well.

    Note
    ----
    Return None if `patterns` is empty or not set.

    Examples
    --------
    Basic usage examples

    >>> _match("README.md", []) is None
    True
    >>> _match("README.md", ["*.txt", "*.yml"]) is None
    True
    >>> _match(".trufflehog.yml", ["*.txt", "*.yml"])
    '*.yml'
    >>> _match(".git/hooks/update.sample", [".git/*"]) is None
    True
    >>> _match(".git/hooks/update.sample", [".git/*"], recursive=True)
    '.git/*'

    """
    if not patterns:
        return None

    fpath = Path(path)
    paths = [fpath, *fpath.parents] if recursive else [fpath]
    for p in paths:
        for glob in patterns:
            if p.match(glob):
                return glob

    return None



"""Run trufflehog3 as a module.

Currently, it is the same as running it with CLI

```bash
$ python -m trufflehog3 --help
```
"""

from trufflehog3 import cli  # pragma: no cover

cli.run()  # pragma: no cover



"""Supported search algorithms."""

from typing import Iterable, Iterator, Optional, Union

from trufflehog3 import NOSECRET_INLINE_RE, IGNORE_NOSECRET
from trufflehog3 import helper, log, source
from trufflehog3.models import Entropy, Exclude, File, Issue, Pattern

MATCH_ALL_RULE_IDS = "*"


def search(
    file: File,
    rules: Iterable[Union[Entropy, Pattern]],
    exclude: Iterable[Exclude] = None,
    ignore_nosecret: bool = IGNORE_NOSECRET,
    context: int = 0,
) -> Iterable[Issue]:
    """Return issues found using provided rules.

    Examples
    --------
    Entropy-based search

    >>> rule = Entropy()
    >>> file = File(
    ...     path="/path/to/code.py",
    ...     content="password = 'abcdefghijklmnopqrstuvwxyz'",
    ... )
    >>> for issue in search(file, [rule]):
    ...     print(issue.secret)
    abcdefghijklmnopqrstuvwxyz

    Pattern-based search

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="letmein",
    ...     severity="high",
    ... )
    >>> file = File(
    ...     path="/path/to/code.py",
    ...     content="password = 'letmein'",
    ... )
    >>> for issue in search(file, [rule]):
    ...     print(issue.secret)
    letmein

    With exclude

    >>> exc = Exclude(message="Not a secret", paths=["*.py"])
    >>> len(search(file, [rule], exclude=[exc]))
    0

    With inline exclude

    >>> file = File(
    ...     path="/path/to/code.py",
    ...     content="password = 'letmein'  # nosecret",
    ... )
    >>> len(search(file, [rule]))
    0

    >>> file = File(
    ...     path="/path/to/code.py",
    ...     content="password = 'letmein'  # nosecret: bad-password-letmein",
    ... )
    >>> len(search(file, [rule]))
    0

    """
    return list(searchiter(file, rules, exclude, ignore_nosecret, context))


def searchiter(
    file: File,
    rules: Iterable[Union[Entropy, Pattern]],
    exclude: Iterable[Exclude] = None,
    ignore_nosecret: bool = IGNORE_NOSECRET,
    context: int = 0,
) -> Iterator[Issue]:
    """Yield issues found using provided rules."""
    content = file.read()

    for i, line in enumerate(content.splitlines()):
        line_number = i + 1
        exclude_ids = [] if ignore_nosecret else _parse_nosecret(line)
        location = f"{file.path}:{line_number}"

        if MATCH_ALL_RULE_IDS in exclude_ids:
            log.info(f"nosecret: skipping {location}")
            continue

        for rule in rules:
            if rule.id.lower() in exclude_ids:
                log.info(f"nosecret: skipping {rule.id} in {location}")
                continue

            for match in rule.findall(line):
                issue = Issue(
                    rule=rule,
                    path=file.path,
                    line=str(line_number),
                    secret=match,
                    context=helper.get_lines(content, line_number, context),
                    branch=file.branch,
                    message=file.message,
                    author=file.author,
                    commit=file.commit,
                    date=file.date,
                )

                if _match(issue, exclude):
                    log.info(f"exclude: skipping {rule.id} in {location}")
                    continue

                yield issue


def _parse_nosecret(s: str) -> Iterable[str]:
    """Parse `nosecret` comment from string and return excluded rule IDs.

    Note
    ----
    Return `'*'` wildcard if no rule IDs were set in the comment.

    Examples
    --------
    Basic usage examples

    >>> _parse_nosecret("token = get_token()  # not secret")
    []
    >>> _parse_nosecret("token = get_token()  # nosecret")
    ['*']
    >>> _parse_nosecret("token = get_token()  # nosecret: rule-id")
    ['rule-id']
    >>> sorted(_parse_nosecret("token := getToken() // nosecret: id1,id2"))
    ['id1', 'id2']
    >>> sorted(_parse_nosecret("token := getToken() // NOSECRET:id1, id2"))
    ['id1', 'id2']

    """
    match = NOSECRET_INLINE_RE.search(s)
    if match is None:
        return []

    ids = match.groupdict()["ids"]
    if ids is None:
        return [MATCH_ALL_RULE_IDS]

    return list(set(id.strip().lower() for id in ids.split(",") if id.strip()))


def _match(
    issue: Issue, exclude: Iterable[Exclude] = None
) -> Optional[Exclude]:
    r"""Match issue against given exclude rules and return matched rule if any.

    Note
    ----
    Return None if `exclude` is not set.

    Examples
    --------
    Basic usage examples

    >>> rule = Pattern(
    ...     id="bad-password-letmein",
    ...     message="Bad Password 'letmein'",
    ...     pattern="letmein",
    ...     severity="high",
    ... )
    >>> issue = Issue(
    ...     rule=rule,
    ...     path="/path/to/code.py",
    ...     line="10",
    ...     secret="letmein",
    ...     context={
    ...         "9":  "username = 'admin'",
    ...         "10": "password = 'letmein'",
    ...         "11": "response = authorize(username, password)",
    ...     },
    ... )
    >>> erule = Exclude(message="Not a secret", id="bad-password-notpass")
    >>> _match(issue, [erule]) is None
    True
    >>> erule = Exclude(message="Not a secret", id="bad-password-letmein")
    >>> _match(issue, [erule]).message
    'Not a secret'
    >>> erule = Exclude(message="Not a secret", paths=["/path/to/*"])
    >>> _match(issue, [erule]).message
    'Not a secret'
    >>> erule = Exclude(message="Not a secret", pattern="letmein")
    >>> _match(issue, [erule]).message
    'Not a secret'
    >>> erule = Exclude(message="Not a secret", pattern="(?i)letmein")
    >>> _match(issue, [erule]).message
    'Not a secret'

    """
    if exclude is None:
        return None

    for exc in exclude:
        if exc.paths:
            path_matched = source._match(issue.path, exc.paths, recursive=True)
        else:
            path_matched = True

        if path_matched and (
            not any((exc.id, exc.pattern))
            or (
                exc.id == issue.rule.id
                or exc.findall(issue.context[issue.line])
            )
        ):
            return exc

    return None
